# 蛋壳创意科技--技术分享--缓存

## 缓存的概念

### 外存

> ​		外储存器是指除计算机内存及CPU 缓存以外的储存器，此类储存器一般断电后仍然能保存数据。
> 常见的外存储器有硬盘、软盘、光盘、U 盘等，一般的软件都是安装在外存中（windows系统指的是CDEF 盘, Linux 系统指的是挂载点）。

### 内存

> ​		内存是计算机中重要的部件之一，它是与CPU 进行沟通的桥梁。计算机中所有程序的
> 运行都是在内存中进行的，因此内存的性能对计算机的影响非常大。
> ​		内存(Memory)也被称为内存储器，其作用是用于暂时存放CPU 中的运算数据，以及与硬盘等外部存储器交换的数据。只要计算机在运行中，CPU 就会把需要运算的数据调到内存中进行运算，当运算完成后CPU 再将结果传送出来，内存的运行也决定了计算机的稳定运行，此类储存器一般断电后数据就会被清空。

### 缓存

> ​		缓存就是把一些外存上的数据保存到内存上而已，怎么保存到内存上呢，我们运行的所有程序，里面的变量值都是放在内存上的，所以说如果要想使一个值放到内存上，实质就是在获得这个变量之后，用一个生存期较长的变量存放你想存放的值，在java中一些缓存一般都是通过map 集合来做的。
>
> ​		广义的缓存是把一些慢存（较慢的外存）上的数据保存到快存（较快的存储）上，简单讲就是，如果某些资源或者数据会被频繁的使用，而这些资源或数据存储在系统外部，比如数据库、硬盘文件等，那么每次操作这些数据的时候都从数据库或者硬盘上去获取，速度会很慢，会造成性能问题（系统停工待料）。于是我们把这些数据冗余一份到快存里面，每次操作的时候，先到快存里面找，看有没有这些数据，如果有，那么就直接使用，如果没有那么就获取它，并复制一份到快存中，下一次访问的时候就可以直接从快存中获取。从而节省大量的时间，可以看出，缓存是一种典型的空间换时间的方案。
>
> > 生活中这样的例子处处可见，举例：
> >
> > - 举例1 --CPU--L1/L2--内存--磁盘
> >   CPU 需要数据时先从L1/L2 中读取，如果没有到内存中找，如果还没有会到磁盘上找。
> > -  举例2 --Maven
> >   还有如用过Maven 的朋友都应该知道，我们找依赖的时候，先从本机仓库找，再从本地服务器仓库找，最后到远程仓库服务器找。
> > - 举例3 --京东仓储
> >   还有如京东的物流为什么那么快？他们在各个地都有分仓库，如果该仓库有货物那么送货的速度是非常快的。
> > -  举例4 --数据库中的索引，也是以空间换取时间，缓存也是以空间换取时间。

#### 缓存的重要指标

##### 缓存命中率

> 表明缓存是否运行良好的
> 即【从缓存中读取数据的次数】与【总读取次数】的比率，命中率越高越好：
>
> ###### 缓存命中率
>
> > 命中率= `从缓存中读取次数`/ (`总读取次数`[==从缓存中读取次数==+ ==从慢速设备上读取
> > 的次数==])
>
> ###### 缓存MISS率
>
> > Miss 率= `没有从缓存中读取的次数`/ (`总读取次数`[==从缓存中读取次数==+ ==从慢速设备
> > 上读取的次数==])
>
> 这是一个非常重要的监控指标，如果做缓存一定要健康这个指标来看缓存是否工作良好

##### 缓存移除策略

> 不同的移除策略实际上看的是不同的指标
> 即如果缓存满了，从缓存中移除数据的策略；常见的有LFU、LRU、FIFO：
>
> ###### FIFO : `First In First Out`
>
> > 先进先出算法，即先放入缓存的先被移除
>
> ###### LRU: `Least Recently Used`
>
> > 久未使用算法，使用时间距离现在最久的那个被移除
>
> ###### LFU: `Least Frequently Used`
>
> > 最近最少使用算法，一定时间段内使用次数（频率）最少的那个被移除
>
> ###### ==TTL: `Time To Live`==
>
> > 存活期，即从缓存中创建时间点开始直到它到期的一个时间段（不管在这个时间段内有没有访问都将过期）
>
> ###### TTI：Time To Idle
>
> > 空闲期，即一个数据多久没被访问将从缓存中移除的时间。
>
> <b>实际设计缓存时以上重要指标都应该考虑进去，当然根据实际需求可能有的指标并不会采用进设计去</b>

## 缓存的逻辑处理流程

![](https://img.mercymodest.com/public/20210326145533.png)

### 代码示例

language：`java`

cache: `redis`

```java
    @Override
    public Optional<Provinces> detailWithRedisCache(String provincesId) {
        if (StrUtil.isBlank(provincesId)) {
            return Optional.empty();
        }
        String key = StrUtil.format(Const.RedisKey.KEY_PATTERN_PROVINCES_DETAIL, provincesId);
        String cache = stringRedisTemplate.opsForValue().get(key);
        if (StrUtil.isNotBlank(cache)) {
            if (log.isDebugEnabled()) {
                log.debug("detailWithRedisCache  ==> {} success hit cache {}", Thread.currentThread().getName(), cache);
            }
            // 缓存命中
            return gson.fromJson(cache, new TypeToken<Optional<Provinces>>() {
            }.getType());
        } else {
            // 数据没有缓存  
            // TODO 缓存击穿 分布式锁  etc. Redisson 可以参阅 cn.dankal.share.cache.reference
            Provinces provinces = this.getProvincesByProvinces(provincesId);
            Optional<Provinces> provincesOptional;
            if (BeanUtil.isNotEmpty(provinces)) {
                if (log.isDebugEnabled()) {
                    log.debug("detailWithRedisCache  ==> {} success get provinces by {} from  database.", Thread.currentThread().getName(), provincesId);
                }
                provincesOptional = Optional.of(provinces);
                final long timeoutInMinutes = Const.CacheTimeout.TIMEOUT_PROVINCES_DETAIL_MINUTES;
                stringRedisTemplate.opsForValue().set(key, gson.toJson(provincesOptional), timeoutInMinutes, TimeUnit.MINUTES);
            } else {
                // 防止缓存穿透 
                if (log.isDebugEnabled()) {
                    log.debug("detailWithRedisCache  ==> {} preven cache break down .", Thread.currentThread().getName());
                }
                final long timeoutInMinutes = Const.CacheTimeout.TIMEOUT_PROVINCES_DETAIL_MINUTES;
                provincesOptional = Optional.empty();
                stringRedisTemplate.opsForValue()
                        .set(key, gson.toJson(provincesOptional), timeoutInMinutes, TimeUnit.MINUTES);
            }
            // 防止缓存雪崩
            // 每一个缓存 30秒 以内的随机差距 
            Long expire = stringRedisTemplate.getExpire(key, TimeUnit.SECONDS);
            expire += RandomUtil.randomLong(30);
            stringRedisTemplate.expire(key, expire, TimeUnit.SECONDS);
            return provincesOptional;
        }
    }
```

## Spring Cache

![](https://img.mercymodest.com/public/20210326174415.png)

### `@Cacheable`

> ![](https://img.mercymodest.com/public/20210326174521.png)

### `@CachePut`

> ![](https://img.mercymodest.com/public/20210326174556.png)

### `@CacheEivct`

> ![](https://img.mercymodest.com/public/20210326174630.png)

![](https://img.mercymodest.com/public/20210326174746.png)

### 举个小栗子

#### 使用`spring cache` 的两个前提和一个自定义

##### 前提一: 开启缓存

`@EnableCaching`

![](https://img.mercymodest.com/public/20210326175249.png)

##### 前提二: 注册 `CacheManager`

`org.springframework.cache.CacheManager`

etc.

```java
 /**
     * 注册一个 {@link CacheManager}
     *
     * @param connectionFactory {@link RedisConnectionFactory}
     * @return  {@link CacheManager}
     */
    @Bean
    public CacheManager cacheManager(RedisConnectionFactory connectionFactory) {
        return RedisCacheManager
                .builder(connectionFactory)
                .cacheDefaults(
                        RedisCacheConfiguration.defaultCacheConfig()
                                //缓存时间绝对过期时间120s
                                .entryTtl(Duration.ofSeconds(120)))
                .transactionAware()
                .build();
    }
```

##### 可选自定义: 缓存 `key` ：`KeyGenerator`

`org.springframework.cache.interceptor.KeyGenerator`

etc.

tips: Spring Cache 的默认 KeyGenerator 默认也是这么做滴

```java
 /**
     * {@link Cacheable} 默认 {@link KeyGenerator}
     *
     * @return {@link KeyGenerator}
     */
    @Bean("customerKeyGenerator")
    public KeyGenerator keyGenerator() {
        return (Object target, Method method, Object... params) -> {
            StringBuilder stringBuilder = new StringBuilder();
            for (int i = 0; i < params.length; i++) {
                stringBuilder.append(params[i]);
            }
            return stringBuilder.toString();
        };
    }
```

#### 缓存方法

```java
   @Cacheable(cacheNames = "provinces_detail_cache", keyGenerator = "customerKeyGenerator")
    @Override
    public Optional<Provinces> detailWithSpringCache(String provincesId) {
        Provinces provinces = this.getProvincesByProvinces(provincesId);
        return Optional.ofNullable(provinces);
    }
```

#### 缓存更新方法: 直接删除缓存

```java
    @CacheEvict(cacheNames = "provinces_detail_cache", key = "#provinces.provinceId")
    @Override
    public Optional<Integer> updateProvinces(Provinces provinces) {
        LambdaQueryWrapper<Provinces> provincesLambdaQueryWrapper = new LambdaQueryWrapper<Provinces>()
                .eq(Provinces::getProvinceId, provinces.getProvinceId());
        int update = baseMapper.update(provinces, provincesLambdaQueryWrapper);
        return Optional.of(update);
    }
```

#### 效果演示

url: GET http://localhost:9102/test/440000

##### 调用 缓存方法

![](https://img.mercymodest.com/public/20210326180109.png)

##### 调用 更新缓存方法

![](https://img.mercymodest.com/public/20210326180401.png)

## 缓存过期与数据一致性问题

![](https://img.mercymodest.com/public/20210326174900.png)

![](https://img.mercymodest.com/public/20210326174939.png)

## 缓存击穿

> ​	 在平常高并发的系统中，大量的请求同时查询一个 key 时，此时这个key正好失效了，就会导致大量的请求都打到数据库上面去。这种现象我们称为**缓存击穿**

![](https://img.mercymodest.com/public/20210326182420.png)

### 解决方案示例

- 分布式锁： etc. `Redis 分布式锁` `Zookeepr 分布式锁` `Redisson`

## 缓存穿透

> ​     那么请求去查询一条压根儿数据库中根本就不存在的数据，也就是缓存和数据库都查询不到这条数据，但是请求每次都会打到数据库上面去。
>
> 这种查询不存在数据的现象我们称为**缓存穿透**。

![](https://img.mercymodest.com/public/20210326182212.png)

布隆过滤器博文参考：[Bloom Filter](https://www.cnblogs.com/z941030/p/9218356.html)

#### 小示例

##### 第三方依赖

> [hutool-bloomFilter](https://mvnrepository.com/artifact/cn.hutool/hutool-bloomFilter)

##### 数据初始化

```java
    @PostConstruct
    public void postConstruct() {
        bitMapBloomFilter = new BitMapBloomFilter(10);
        LambdaQueryWrapper<Provinces> provincesLambdaQueryWrapper = new LambdaQueryWrapper<Provinces>()
                .select(Provinces::getProvinceId);
        baseMapper.selectList(provincesLambdaQueryWrapper)
                .stream()
                .filter(Objects::nonNull)
                .map(Provinces::getProvinceId)
                .forEach(bitMapBloomFilter::add);
    }
```

##### 使用 `BitMapBloomFilter` 防止缓存穿透

```java
   @Override
    public Optional<Provinces> detailsWithBoolmFilter(String provinceId) {
        if (StrUtil.isBlank(provinceId)) {
            return Optional.empty();
        }
        if (!bitMapBloomFilter.contains(provinceId)) {
            // provinceId 不存在
            return Optional.empty();
        }
        return Optional.ofNullable(this.getProvincesByProvinces(provinceId));
    }
```

##### 小提示

> 示例 只是单纯的对 `BitMapBloomFilter`进行初数化，我们还需考虑其数据更新.我们可以在 Provinces进行增删的时候手动对 `BitMapBloomFilter`进行更新。当然我们也可以通过订阅 诸如`TiDB`的`Ticdc`来实现`BitMapBloomFilter`的数据更新

### 解决方案示例

- 缓存空值

- BloomFilter ： 布隆过滤器

  ![](https://img.mercymodest.com/public/20210326182241.png)

## 缓存雪崩

> ​	缓存雪崩的情况是说，当某一时刻发生大规模的缓存失效的情况，比如你的缓存服务宕机了，会有大量的请求进来直接打到DB上面。结果就是DB 称不住，挂掉。

![](https://img.mercymodest.com/public/20210326182354.png)

### 解决方案示例

- 使用集群缓存，保证缓存服务的高可用
- 对应用服务进行限流
- 缓存数据的时候，在原来过期时间的基础上加上一定范围的随机数

