package cn.dankal.share.cache.config;

import org.springframework.cache.CacheManager;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.cache.interceptor.KeyGenerator;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.data.redis.cache.RedisCacheConfiguration;
import org.springframework.data.redis.cache.RedisCacheManager;
import org.springframework.data.redis.connection.RedisConnectionFactory;

import java.lang.reflect.Method;
import java.time.Duration;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-03-26
 */
@Configuration
public class SpringCacheConfig {

    /**
     * {@link Cacheable} 默认 {@link KeyGenerator}
     *
     * @return {@link KeyGenerator}
     */
    @Bean("customerKeyGenerator")
    public KeyGenerator keyGenerator() {
        return (Object target, Method method, Object... params) -> {
            StringBuilder stringBuilder = new StringBuilder();
            for (int i = 0; i < params.length; i++) {
                stringBuilder.append(params[i]);
            }
            return stringBuilder.toString();
        };
    }

    /**
     * 注册一个 {@link CacheManager}
     *
     * @param connectionFactory {@link RedisConnectionFactory}
     * @return  {@link CacheManager}
     */
    @Bean
    public CacheManager cacheManager(RedisConnectionFactory connectionFactory) {
        return RedisCacheManager
                .builder(connectionFactory)
                .cacheDefaults(
                        RedisCacheConfiguration.defaultCacheConfig()
                                //缓存时间绝对过期时间120s
                                .entryTtl(Duration.ofSeconds(120)))
                .transactionAware()
                .build();
    }
}
