package cn.dankal.share.cache.service;

import cn.dankal.share.cache.config.SpringCacheConfig;
import cn.dankal.share.cache.entity.Provinces;
import cn.hutool.bloomfilter.BitMapBloomFilter;
import com.baomidou.mybatisplus.extension.service.IService;
import org.springframework.cache.annotation.CacheEvict;

import java.util.Optional;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-03-26
 */
public interface IProvincesService extends IService<Provinces> {

    /**
     * 获取身份详情(无缓存方案)
     *
     * @param provincesId 身份 id
     * @return {@link Optional}  contains {@link Provinces}
     */
    Optional<Provinces> detailNoCache(String provincesId);

    /**
     * 获取身份详情(redis 缓存方案)
     *
     * @param provincesId 身份 id
     * @return {@link Optional}  contains {@link Provinces}
     */
    Optional<Provinces> detailWithRedisCache(String provincesId);

    /**
     * 更新 {@link Provinces}
     * {@link CacheEvict} 删除缓存
     *
     * @param provinces provinces
     * @return {@link Optional}  contains {@link Integer}
     */
    Optional<Integer> updateProvinces(Provinces provinces);

    /**
     * 获取身份详情(spring cache 缓存方案)
     * 使用 {@link SpringCacheConfig#keyGenerator()} 作为 缓存 key 缓存 方法执行结果
     *
     * @param provincesId 身份 id
     * @return {@link Optional}  contains {@link Provinces}
     */
    Optional<Provinces> detailWithSpringCache(String provincesId);


    /**
     * 测试: 使用 {@link BitMapBloomFilter} 防止缓存击穿 (伪代码实现)
     *
     * @param provinceId provincesId 身份 id
     * @return {@link Optional}  contains {@link Provinces}
     */
    Optional<Provinces> detailsWithBoolmFilter(String provinceId);
}
